#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/interrupt.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <linux/input.h>

#include <asm/uaccess.h>
#include <asm/irq.h>
#include <asm/io.h>

#define AUTHOR		"Atmark Techno, Inc."
#define DESCRIPTION	"Armadillo-2x0 Tact-SW driver"
#define LICENSE		"GPL v2"
#define REVISION	(CONFIG_INPUT_ARMADILLO2X0_TACTSW_REV)

#define DRIVER_NAME	"tactsw"
#define DRIVER_DEVICE	(DRIVER_NAME)

#if defined(DEBUG)
#define KDEBUG_FUNC()         printk(DRIVER_NAME ": %s()\n", __FUNCTION__)
#define KDEBUG_INFO(args...)  printk(DRIVER_NAME ": " args)
#define KDEBUG_ERR_LINE()     printk(DRIVER_NAME ": Error(%d)\n", __LINE__)
#else
#define KDEBUG_FUNC()
#define KDEBUG_INFO(args...)
#define KDEBUG_ERR_LINE()
#endif
#define PRINT_INFO(args...)  printk(KERN_INFO  DRIVER_NAME ": " args)
#define PRINT_DEBUG(args...) printk(KERN_DEBUG DRIVER_NAME ": " args)
#define PRINT_ERR(args...)   printk(KERN_ERR   DRIVER_NAME ": " args)

struct gpio_port_reg {
	unsigned long ddr;
	unsigned long dr;
	unsigned long inten;
	unsigned long inttype1;
	unsigned long inttype2;
	unsigned long eoi;
	unsigned long db;
	unsigned long rawintsts;
	unsigned long intsts;
};

static struct gpio_port_reg portf = {
	.ddr       = GPIO_PFDDR,
	.dr        = GPIO_PFDR,
	.inten     = GPIO_FINTEN,
	.inttype1  = GPIO_FINTTYPE1,
	.inttype2  = GPIO_FINTTYPE2,
	.eoi       = GPIO_FEOI,
	.db        = GPIO_FDB,
	.rawintsts = GPIO_RAWINTSTATUSF,
	.intsts    = GPIO_INTSTATUSF,
};

struct armadillo2x0_tactsw_port {
	char *name;
	struct gpio_port_reg *gpio;
	unsigned long bit;

	unsigned int irq;

	struct input_dev idev;
};

static struct armadillo2x0_tactsw_port ports[] = {
	{ "Tact-SW Port1", &portf, BIT(7), IRQ_GPIO7 },
};

static int
get_idev_index(struct input_dev *idev)
{
	int i;
	for (i=0; i<ARRAY_SIZE(ports); i++) {
		if (idev == &ports[i].idev)
			return i;
	}
	return -1;
}

static int
tactsw_enable_irq(struct input_dev *idev)
{
	int port = get_idev_index(idev);
	writel(readl(ports[port].gpio->inten) | ports[port].bit,
	       ports[port].gpio->inten);
	return 0;
}

static int
tactsw_disable_irq(struct input_dev *idev)
{
	int port = get_idev_index(idev);
	writel(readl(ports[port].gpio->inten) & ~ports[port].bit,
	       ports[port].gpio->inten);
	return 0;
}

static int
get_port_status(struct input_dev *idev)
{
	int port = get_idev_index(idev);

	return readl(ports[port].gpio->dr) & ports[port].bit ? 1 : 0;
}

static int
armadillo2x0_tactsw_open(struct input_dev *idev)
{
	int port = get_idev_index(idev);
	KDEBUG_FUNC();

	tactsw_disable_irq(idev);

	if (get_port_status(idev))
		/* change: Falling-Edge */
		writel(readl(ports[port].gpio->inttype2) & ~ports[port].bit,
		       ports[port].gpio->inttype2);
	else
		/* change: Riging-Edge */
		writel(readl(ports[port].gpio->inttype2) | ports[port].bit,
		       ports[port].gpio->inttype2);

	tactsw_enable_irq(idev);

	return 0;
}

static void
armadillo2x0_tactsw_close(struct input_dev *idev)
{
	KDEBUG_FUNC();

	tactsw_disable_irq(idev);
}

static irqreturn_t
armadillo2x0_tactsw_irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	struct input_dev *idev = (struct input_dev *)dev_id;
	int port = get_idev_index(idev);

	KDEBUG_FUNC();

	if (!(readl(ports[port].gpio->intsts) & ports[port].bit))
		return IRQ_NONE;

	input_regs(idev, regs);

	if (get_port_status(idev)) {
		/* change: Falling-Edge */
		writel(readl(ports[port].gpio->inttype2) & ~ports[port].bit,
		       ports[port].gpio->inttype2);
		input_event(idev, EV_SW, 0, 0); /* Up! */
	} else {
		/* change: Riging-Edge */
		writel(readl(ports[port].gpio->inttype2) | ports[port].bit,
		       ports[port].gpio->inttype2);
		input_event(idev, EV_SW, 0, 1); /* Down! */
	}

	/* ACK */
	writel(ports[port].bit, ports[port].gpio->eoi);

	input_sync(idev);

	return IRQ_HANDLED;
}

static int
armadillo2x0_tactsw_ports_setup(void)
{
	int i;

	for (i=0; i<ARRAY_SIZE(ports); i++) {
		writel(readl(ports[i].gpio->dr) & ~ports[i].bit,
		       ports[i].gpio->dr);
		writel(readl(ports[i].gpio->ddr) & ~ports[i].bit,
		       ports[i].gpio->ddr);
		/* Default: Falling-Edge */
		writel(readl(ports[i].gpio->inttype1) | ports[i].bit,
		       ports[i].gpio->inttype1);
		writel(readl(ports[i].gpio->inttype2) & ~ports[i].bit,
		       ports[i].gpio->inttype2);
		writel(readl(ports[i].gpio->db) | ports[i].bit,
		       ports[i].gpio->db);
		writel(readl(ports[i].gpio->inten) | ports[i].bit,
		       ports[i].gpio->inten);
	}

	return 0;
}

static int
armadillo2x0_tactsw_ports_release(void)
{
	return 0;
}

static int
armadillo2x0_tactsw_probe(void)
{
	int ret;
	int i;

	KDEBUG_FUNC();

	ret = armadillo2x0_tactsw_ports_setup();
	if (ret)
		goto _err_tactsw_ports_setup;

	for (i=0; i<ARRAY_SIZE(ports); i++) {
		ret = request_irq(ports[i].irq,
				  armadillo2x0_tactsw_irq_handler,
				  SA_INTERRUPT, ports[i].name, &ports[i].idev);
		if (ret < 0) {
			while (i-- >= 0)
				free_irq(ports[i].irq, &ports[i].idev);
			goto _err_request_irq;
		}
		tactsw_disable_irq(&ports[i].idev);
	}

	for (i=0; i<ARRAY_SIZE(ports); i++) {
		ports[i].idev.name = ports[i].name;
		ports[i].idev.phys = NULL;
		ports[i].idev.id.bustype = BUS_HOST;

		ports[i].idev.open = armadillo2x0_tactsw_open;
		ports[i].idev.close = armadillo2x0_tactsw_close;

		ports[i].idev.evbit[0] = BIT(EV_SW);
		ports[i].idev.swbit[0] = BIT(0);

		input_register_device(&ports[i].idev);
	}

	return 0;

 _err_request_irq:
	armadillo2x0_tactsw_ports_release();
 _err_tactsw_ports_setup:

	printk("Unable to register %s driver.\n", DRIVER_NAME);
	return ret;
}

static int
armadillo2x0_tactsw_remove(void)
{
	int i;

	KDEBUG_FUNC();

	for (i=0; i<ARRAY_SIZE(ports); i++) {
		input_unregister_device(&ports[i].idev);
		free_irq(ports[i].irq, &ports[i].idev);
	}

	armadillo2x0_tactsw_ports_release();

	return 0;
}

static void
print_driver_info(void)
{
	printk(KERN_INFO "%s: %s [Rev.%d.%02d]\n",
	       DRIVER_NAME, DESCRIPTION,
	       REVISION/100, REVISION%100);
}

static int __init
armadillo2x0_tactsw_init(void)
{
	KDEBUG_FUNC();

	print_driver_info();

	return armadillo2x0_tactsw_probe();
}

static void __exit
armadillo2x0_tactsw_exit(void)
{
	KDEBUG_FUNC();
	armadillo2x0_tactsw_remove();
}

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESCRIPTION);
MODULE_LICENSE(LICENSE);

module_init(armadillo2x0_tactsw_init);
module_exit(armadillo2x0_tactsw_exit);
